#include "MyGLWidget.hpp"

#include <fstream>
#include <iostream>

using namespace cat;

MyGLWidget::MyGLWidget(QWidget *parent)
    : QGLWidget(parent)
{
    startTimer(15); // shooting for 60 FPS
}

MyGLWidget::~MyGLWidget()
{
}

QSize MyGLWidget::minimumSizeHint() const
{
    return QSize(640, 640);
}

QSize MyGLWidget::sizeHint() const
{
    return QSize(800, 800);
}

void MyGLWidget::initializeGL()
{
    // Orthographic projection
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(-3, 3, 3, -3, -15, 15);

    // Enable smooth color gradients between vertices
    glShadeModel(GL_SMOOTH);

    // Identity modelview
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    // Background color is black
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    // Enable the depth test to cause faces to render from back to front
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // Use the nicest perspective correction hint
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // Enable cull face so faces with normals pointing away from the camera
    // do not get rendered
    glEnable(GL_CULL_FACE);

    angle = 0.0f;
    scale = 1.f;
    vscale = 0.1f;
    zoom = 1.f;
    vzoom = 0.01f;
    frame = 0;

    for (int ii = 0; ii < 100; ++ii)
    {
        precalc_xy[ii] = (float)ii*6.f/100.f - 3.f;
    }

    for (int ii = 0; ii < 100; ++ii)
    {
        for (int jj = 0; jj < 100; ++jj)
        {
            float x = precalc_xy[ii];
            float y = precalc_xy[jj];
            precalc_z[ii][jj] = sqrt(x * x + y * y);
        }
    }

    // Incredibly small monochrome bitmap loader
    std::ifstream file("mask.bmp", std::ios_base::binary);
    if (file)
    {
        file.seekg(0x3d, std::ios_base::beg);

        const int row_bytes = ((100 + 31) / 32) * 4;
        char row[row_bytes];

        for (int jj = 0; jj < 100; ++jj)
        {
            file.read(row, row_bytes);

            for (int ii = 0; ii < 100; ++ii)
            {
                int kk = ii + 8;
                mask[ii][jj] = (row[kk >> 3] >> (7 - (kk & 7))) & 1;
            }
        }
    }

    // Quaternion camera
    camera_rotation.setFromAxisAngle(Vector3d(1, 0, 0), Deg2Rad(30.f));
    camera_rotation_interval.setFromAxisAngle(Vector3d(0, 0, 1), Deg2Rad(1.f));
}

void MyGLWidget::paintGL()
{
    // Clear color (background) and depth (for depth test) buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);

    camera_rotation.postMultiply(camera_rotation_interval);

    // Quaternion camera rotation
    Matrix4x4d m;
    camera_rotation.getMatrix4x4(m);
    glLoadMatrixd(&m(0));

    glScalef(zoom, zoom, zoom);

    cube.begin_draw();
    //sphere.draw();
    //sphere.end_draw();
    //return;

    for (int ii = 0; ii < 100; ++ii)
    {
        float x = precalc_xy[ii];

        for (int jj = 0; jj < 100; ++jj)
        {
            float y = precalc_xy[jj];
            float m = precalc_z[ii][jj];
            float z = (float)cos(m * scale) / (-m);

            glPushMatrix();
            glTranslatef(x, y, z);
            if (mask[(jj + frame) % 100][ii])
                glRotatef(180.f, 1, 0, 0);
            cube.draw();
            glPopMatrix();
        }
    }

    cube.end_draw();
}

void MyGLWidget::timerEvent(QTimerEvent *)
{
    ++frame;

    angle += 0.3f; // degrees

    // Scale modulation
    scale += vscale;

    if (scale > 10.f)
    {
        vscale = -.1f;
    }
    else if (scale < 0.f)
    {
        vscale = .1f;
    }

    // Zoom
    zoom += vzoom;

    if (zoom > 3.f)
    {
        vzoom = -.01f;
    }
    else if (zoom < 0.5f)
    {
        vzoom = .01f;
    }

    // Redraw on a timer
    update();
}

void MyGLWidget::resizeGL(int width, int height)
{
    // Fix viewport to new size
    glViewport(0, 0, width, height);

    // Letter-boxed version
    glEnable(GL_SCISSOR_TEST);
    glScissor(0, height * 1 / 8, width, height * 3 / 4);
}

void MyGLWidget::mousePressEvent(QMouseEvent *event)
{
    close();
}

void MyGLWidget::mouseMoveEvent(QMouseEvent *event)
{
    close();
}

void MyGLWidget::keyPressEvent(QKeyEvent *event)
{
    close();
}
