#include "Sphere.hpp"

#include <cat/AllGraphics.hpp>
using namespace cat;

// Ctor
Sphere::Sphere()
{
    _valid = false;
}

// Dtor
Sphere::~Sphere()
{
    clean();
}

// Initializing the vertex buffer objects
void Sphere::init()
{
    // Spherical geometry from a normalized cube

    // Generate one side of the cube:

#define DERIVATIVE(x, y, u0, u1, u2) \
    { \
        GLfloat *u = &_vertices[(y * R + x) * 3]; \
        u[0] = u0; u[1] = u1; u[2] = u2; \
    }

    f32 z = (R/2-1) + .5f;

    // Trace out an eighth triangle of the geometry, which is linearly related to the rest
    for (int iy = 0; iy < S; ++iy)
    {
        f32 y = iy + .5f;

        for (int ix = 0; ix <= iy; ++ix)
        {
            f32 x = ix + .5f;

            // Normalize the vector (x,y,z)
            f32 inv_mag = InvSqrt(x*x + y*y + z*z);
            f32 vx = x * inv_mag;
            f32 vy = y * inv_mag;
            f32 vz = z * inv_mag;

            DERIVATIVE(S   + ix, S   + iy, vx, vy, vz); // Natural orientation
            DERIVATIVE(S   + ix, S-1 - iy, vx, -vy, vz); // Flip across y=0
            DERIVATIVE(S-1 - ix, S   + iy, -vx, vy, vz); // Flip across x=0
            DERIVATIVE(S-1 - ix, S-1 - iy, -vx, -vy, vz); // Flip across x=0, y=0

            // When ix == iy, flipping across x=y yields the same four points so skip these:
            if (ix != iy)
            {
                DERIVATIVE(S   + iy, S   + ix, vy, vx, vz); // Flip across x=y
                DERIVATIVE(S   + iy, S-1 - ix, vy, -vx, vz); // Flip across x=y, y=0
                DERIVATIVE(S-1 - iy, S   + ix, -vy, vx, vz); // Flip across x=y, x=0
                DERIVATIVE(S-1 - iy, S-1 - ix, -vy, -vx, vz); // Flip across x=y, x=0, y=0
            }
        }
    }

    // Derive other sides from front face 0
    GLfloat *u = _vertices;
    for (int ii = 0; ii < R*R; ++ii)
    {
        // Top face 1
        u[R*R*3*1] = u[0];
        u[R*R*3*1+1] = -u[2];
        u[R*R*3*1+2] = u[1];

        // Right face 2
        u[R*R*3*2] = u[2];
        u[R*R*3*2+1] = u[0];
        u[R*R*3*2+2] = u[1];

        // Bottom face 3
        u[R*R*3*3] = -u[0];
        u[R*R*3*3+1] = u[2];
        u[R*R*3*3+2] = u[1];

        // Left face 4
        u[R*R*3*4] = -u[2];
        u[R*R*3*4+1] = -u[0];
        u[R*R*3*4+2] = u[1];

        // Back face 5
        u[R*R*3*5] = -u[0];
        u[R*R*3*5+1] = u[1];
        u[R*R*3*5+2] = -u[2];

        u += 3;
    }

    // Generate the index arrays for producing quad strips for each face
    for (int ii = 0; ii < 6; ++ii)
    for (int jj = 0; jj < R-1; ++jj)
    {
        int strip = ii*(R-1) + jj;

        _counts[strip] = R*2;

        int offset = (R*R * ii) + jj;

        for (int kk = 0; kk < R; ++kk)
        {
            _indices[strip][kk*2] = offset + kk*R + 1;
            _indices[strip][kk*2+1] = offset + kk*R;
        }
    }

/*
    if (GLEE_VERSION_1_5)
    {
        // Generate two vertex buffer objects
        glGenBuffers(3, _vbo);

        // Bind buffer 0 to contain the vertex array
        glBindBuffer(GL_ARRAY_BUFFER, _vbo[0]);

        // Load the vertex array into buffer 0
        glBufferData(GL_ARRAY_BUFFER, sizeof(_vertices), _vertices, GL_STATIC_DRAW);

        // Bind buffer 1 to contain the index array
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _vbo[1]);

        // Load the index array into buffer 1
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(_indices), _indices, GL_STATIC_DRAW);

        OGLDIF_CHECK_ERROR;
    }
*/
    _valid = true;
}

// Begin drawing
void Sphere::begin_draw()
{
    if (!_valid)
        init();

    // Push client vertex array state
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

    // Enable using vertex arrays
    glEnableClientState(GL_VERTEX_ARRAY);
/*
    // If OGL version 1.5 or newer,
    if (GLEE_VERSION_1_5)
    {
        // Bind buffer 0 for the vertex array
        glBindBuffer(GL_ARRAY_BUFFER, _vbo[0]);

        // Specify the vertex array from buffer 0
        glVertexPointer(3, GL_FLOAT, 0, bufferObjectPtr(0));

        // Bind buffer 1 for the index array
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _vbo[1]);
    }
    else*/
    {
        // Set the vertex pointer directly
        glVertexPointer(3, GL_FLOAT, 0, _vertices);
    }

    OGLDIF_CHECK_ERROR;
}

// Draw the cube, pulled out from begin/end so a lot of cubes can be drawn quickly
void Sphere::draw()
{
/*    // If OGL version 1.5 or newer,
    if (GLEE_VERSION_1_5)
    {
        glMultiDrawElements(GL_QUAD_STRIP, _counts, GL_UNSIGNED_SHORT, _indices, x);
    }
    else*/
    {
        // If OGL version 1.4 or newer,
/*        if (GLEE_VERSION_1_4)
        {
            glMultiDrawElements(GL_QUAD_STRIP, _counts, GL_UNSIGNED_SHORT, (const GLvoid**)_indices, STRIPS);
        }
        else*/
        {
            // Use glDrawElements() as a fallback
            for (int ii = 0; ii < STRIPS; ++ii)
            {
                glDrawElements(GL_QUAD_STRIP, _counts[ii], GL_UNSIGNED_SHORT, _indices[ii]);
            }
        }
    }
}

// Draw the cube
void Sphere::end_draw()
{
    // Pop client vertex array state
    glPopClientAttrib();

    OGLDIF_CHECK_ERROR;
}

// Clean up
void Sphere::clean()
{
    if (_valid)
    {
        _valid = false;

        if (GLEE_VERSION_1_5)
        {
            glDeleteBuffers(3, _vbo);
        }
    }
}
